/*
 * inputvideo0.c: /dev/video0 input for the Atmolight-plugin
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <vdr/tools.h>
#include "math.h"

#include "defs.h"
#include "setup.h"
#include "input.h"
#include "inputvideo0.h"

// capture width/height
#define CAP_WIDTH    64
#define CAP_HEIGHT   48

// set accuracy of color calculation
#define h_MAX   255
#define s_MAX   255
#define v_MAX   255

// macros
#define MIN(X, Y)  ((X) < (Y) ? (X) : (Y))
#define MAX(X, Y)  ((X) > (Y) ? (X) : (Y))

int d2i(double d)
{
  return (int)(d < 0 ? d - 0.5 : d + 0.5); // TODO: Makro
}

tColorPacket ColorPacket; // TODO: weg, nur interne Bentzung (wenn berhaupt)

// --- cInputVideo0 -----------------------------------------------------------

cInputVideo0::cInputVideo0()
{
  // init variables
  ColorPacket.channel[0].r = ColorPacket.channel[0].g = ColorPacket.channel[0].g = 0.0;
  ColorPacket.channel[4] = ColorPacket.channel[3] = ColorPacket.channel[2] =
                           ColorPacket.channel[1] = ColorPacket.channel[0];
  fd = -1;
  nextframe = 0;
  actframe  = -1;
}

cInputVideo0::~cInputVideo0()
{
}

bool cInputVideo0::Open(const char* param)
{
  int res;

  if ((fd=open("/dev/video0",O_RDWR)) == -1)
  {
    esyslog("Atmolight: Error opening /dev/video0");
    return false;
  }
  // dsyslog("Atmolight: Open /dev/video0");

  // get capabilities and check if capture is possible
  struct video_capability cap;
  ioctl(fd,VIDIOCGCAP,&cap);

  // dsyslog("Atmolight: videosizes: minh=%d maxh=%d minw=%d maxw=%d",cap.minheight,cap.maxheight,cap.minwidth,cap.maxwidth);

  if ((cap.minheight > CAP_HEIGHT) || (cap.maxheight < CAP_HEIGHT) ||  (cap.minwidth > CAP_WIDTH) || (cap.maxwidth < CAP_WIDTH))
  {
    esyslog("Atmolight: videosize not supported!");
    return false;
  }
  // dsyslog("Atmolight: videosize OK");

  // set capture parameters
  struct video_window win;
  res = ioctl(fd,VIDIOCGWIN,&win);
  win.x = win.y = win.chromakey = 0;
  win.width  = CAP_WIDTH;
  win.height = CAP_HEIGHT;
  win.flags = 0;
  if (ioctl(fd,VIDIOCSWIN,&win) == -1)
  {
    esyslog("Atmolight: ioctl: VIDIOCSWIN");
    return false;
  }

  struct video_picture pic;
  res = ioctl(fd,VIDIOCGPICT,&pic);

  // set parameters of captured pictures
  pic.depth = 24;
  pic.palette = VIDEO_PALETTE_RGB24;
  pic.brightness = pic.hue = pic.colour = pic.contrast = pic.whiteness = 32767;
  if (ioctl(fd,VIDIOCSPICT,&pic) == -1)
  {
    esyslog("Atmolight: ioctl: VIDIOCSPICT");
    return false;
  }

  // assign capture buffer
  if (ioctl(fd,VIDIOCGMBUF,&m_buf) == -1)
  {
    perror("ioctl: videodevice VIDIOCGMBUF");
    return false;
  }

#if 0
  if (m_buf.frames < 2)
  {
    esyslog("videodev: cant capture 2 frames!");
    return false;
  }
#endif

  map = mmap(0,m_buf.size,PROT_READ | PROT_WRITE, MAP_SHARED,fd,0);

  if (map == MAP_FAILED)
  {
    esyslog("Atmolight: framebuffer not initialized!");
    return false;
  }

  v_map.frame = 0;
  v_map.width = CAP_WIDTH;
  v_map.height = CAP_HEIGHT;
  v_map.format = VIDEO_PALETTE_RGB24;

  Start(); // start thread

  return true;
}

bool cInputVideo0::Close(void)
{
  Cancel(); // stop thread

  if (map)
  {
    munmap(map, m_buf.size);
  }

  if (fd != -1)
  {
    close(fd);
    fd = -1;
    return true;
  }
  return false;
}

void cInputVideo0::Action(void)
{
  cTimeMs t;

  // cCondWait::SleepMs(1000); // wait 1,0 second // TODO: verkleinern

  while (Running()) // in this loop the picture is read and the colors are calculated
  {
    t.Set(0);

    // read picture and calculate colors
    CalcColors();

    if (t.Elapsed() < 20) // wait 20ms to avoid 100% CPU usage
    {
      cCondWait::SleepMs(20);
    }
  }
}

void* cInputVideo0::Capture(void)
{
  if (fd != -1)
  {
    v_map.frame = nextframe; // set frame to be captured until next

    if (ioctl(fd,VIDIOCMCAPTURE,&v_map) == -1)
    {
      dsyslog("VIDIOCMCAPTURE failed");
      return NULL;
    }

    v_map.frame = actframe; // set frame to read out now

    if (actframe != -1)
    {
      if (ioctl(fd,VIDIOCSYNC,&v_map) == -1)
      {
        dsyslog("VIDIOSYNC failed");
        return NULL;
      }

      if (actframe == 0)
      {
        actframe = 1;
        nextframe = 0;
        return (char*)map;
      }
      else
      {
        nextframe = 1;
        actframe = 0;
        return (char*)map+m_buf.offsets[1];
      }
    }
    else
    {
      actframe = 0;
      nextframe = 1;
      return NULL; // no frame captured yet
    }
  }
  else
  {
    return NULL; // videodev not open
  }
}

void cInputVideo0::CalcColors(void)
{
  unsigned char *map;

  map = (unsigned char *) Capture();

  // handle start problems ;-)
  if (map == NULL) { return; }

  // size of the image
  int Imagesize = CAP_WIDTH * CAP_HEIGHT;

  // generate RGB and HSV image
  tRGBColor *RGB_Img;
  RGB_Img = new tRGBColor[Imagesize];
  memset(RGB_Img, 0, Imagesize * sizeof(RGB_Img));

  tHSVColor *HSV_Img;
  HSV_Img = new tHSVColor[Imagesize];
  memset(HSV_Img, 0, Imagesize * sizeof(HSV_Img));

  // RGB and HSV image builtup
  for ( int idx = 0; idx < Imagesize ; ++idx ) //Walk through map
  {
    // generate RGB image
    RGB_Img[idx].b = (double)*map++ / 255.0f; // -> [0;1]
    RGB_Img[idx].g = (double)*map++ / 255.0f; // -> [0;1]
    RGB_Img[idx].r = (double)*map++ / 255.0f; // -> [0;1]

    // generate HSV image
    HSV_Img[idx] = RGB2HSV(RGB_Img[idx]);
  }

  ColorPacket = CalcColorsAnalyzeHSV(RGB_Img, HSV_Img);

  for (int i = 0; i < 5; i++) // TODO: im Filter mit unsigned int rechnen
  {
    ColorPacketInt.channel[i].r = (unsigned int) (255.0f * ColorPacket.channel[i].r);
    ColorPacketInt.channel[i].g = (unsigned int) (255.0f * ColorPacket.channel[i].g);
    ColorPacketInt.channel[i].b = (unsigned int) (255.0f * ColorPacket.channel[i].b);
  }

  // free memory
  delete[] RGB_Img;
  delete[] HSV_Img;
}

tColorPacket cInputVideo0::CalcColorsAnalyzeHSV(tRGBColor*& RGB_Img, tHSVColor*& HSV_Img)
{
  // for all cases: Channel Index Numbers
  // 0 = center
  // 1 = left
  // 2 = right
  // 3 = top
  // 4 = bottom

  // create weighting matrix

  // size of the image
  int Imagesize = CAP_WIDTH * CAP_HEIGHT;

  tWeightPacket *Weight;
  Weight = new tWeightPacket[Imagesize];
  memset(Weight, 0, Imagesize * sizeof(tWeightPacket));

  int weight_pow_hori = AtmoSetup.WeightPow; // horizontal weighting
  int weight_pow_vert = AtmoSetup.WeightPow; // vertical weighting
  int wdx = 0;
  for (int row = 0; row < CAP_HEIGHT; ++row)
  {
    double row_norm =(double)row/((double)CAP_HEIGHT - 1); // [0;cap->GetCapResHeiht()] -> [0;1]

    // top
    double weight_3 = pow(1.00 - row_norm, weight_pow_vert);
    // bottom
    double weight_4 = pow(row_norm,weight_pow_vert);

    for (int column = 0; column < CAP_WIDTH; ++column)
    {
      // if widescreen mode is set, top and bottom of the picture are not used for the calculations
      if ( (AtmoSetup.WidescreenMode == 1) && ((row <= CAP_HEIGHT/8) || (row >= (7*CAP_HEIGHT)/8)) )
      {
        Weight[wdx].channel[0] = Weight[wdx].channel[1] = Weight[wdx].channel[2] = Weight[wdx].channel[3] = Weight[wdx].channel[4] = 0.0;
      } 
      else
      {
        double column_norm =(double)column / ((double)CAP_WIDTH - 1); // [0;cap->GetCapResWidth()] -> [0;1]
        Weight[wdx].channel[0] = 1.00;         // TODO
        Weight[wdx].channel[1] = (double)pow((1.00 - column_norm), weight_pow_hori);
        Weight[wdx].channel[2] = (double)pow(column_norm, weight_pow_hori);
        Weight[wdx].channel[3] = (double)weight_3;
        Weight[wdx].channel[4] = (double)weight_4;
      }
      wdx++;
    }
  }

  /********************************************************************************/
  /*           Hue                                                                */
  /********************************************************************************/

  /*--------------------------------------*/
  /* Hue Histogram builtup                */
  /*--------------------------------------*/

  int hsv_idx = 0;                            // Histogram Index
  double hue_hist[5][h_MAX+1];                // HSV histogram
  double w_hue_hist[5][h_MAX+1];              // windowed HSV histogram
  memset(&hue_hist, 0, sizeof(hue_hist));     // clean histogram
  memset(&w_hue_hist, 0, sizeof(w_hue_hist)); // clean windowed histogram

  int idx=0;
  for (int row = 0; row < CAP_HEIGHT; ++row)
  {
    for (int column = 0; column < CAP_WIDTH; ++column)
    {
      // forget black bars:
      // perform calculations only if pixel has some saturation and luminosity else
      if (HSV_Img[idx].v > (double)AtmoSetup.DarkLimit * 0.04f)
      {
        hsv_idx = d2i(HSV_Img[idx].h * (double)h_MAX); // convert from [0;1] -> [0;h_MAX] to use as Histogram Index

        // builtup histogram for the 5 channels
        for (int channel = 0; channel < 5; ++channel)
        {
          // Add weight to channel
          hue_hist[channel][hsv_idx] += Weight[idx].channel[channel] * (HSV_Img[idx].v + 1.0f);
        }
      }
      idx++;
    }
  }

  /*--------------------------------------*/
  /* Hue Histogram windowing              */
  /*--------------------------------------*/

  // Windowing over Histogram to get most used Hue
  // TODO implement as function
  // apply lite triangular window design with gradient of 1% per discrete step

  int hue_windowsize = AtmoSetup.HueWinSize; // steps in each direction; eg. 2 => -2 -1 0 1 2 windowing

  // Apply windowing to all 5 channels
  for (int channel = 0; channel < 5; ++channel)
  {
    // Walk through histogram [0;h_MAX]
    for (int h_idx = 0; h_idx < h_MAX+1; ++h_idx)
    {
      // windowing from -hue_windowsize -> +hue_windowsize
      for (int mywin=-hue_windowsize; mywin < hue_windowsize + 1; mywin++ )
      {
        int myidx = h_idx + mywin; // adressed histogram candlestick

        if (myidx < 0) myidx=h_MAX + 1 + myidx; // handle beginning of windowing ->roll back
        if (myidx > h_MAX) myidx=myidx - h_MAX;  // handle end of windowing -> roll forward

        float windesign = 1 - (0.1 * abs(mywin)); // design of window with 10% gradient

        w_hue_hist[channel][h_idx] += (double)(hue_hist[channel][myidx] * windesign); // apply window
        // w_hue_hist[channel][h_idx]=hue_hist[channel][h_idx]; //apply window
      }
    }
  }

  /*--------------------------------------*/
  /* Analyze Histogram for most used Hue  */
  /*--------------------------------------*/

  // Find Max Vaules in windowed Histogram

  double hue_histo_mean[5];  // Mean Value of Histogram
  int hue_dominance[5];      // Calculate faktor from peak over histogram mean
  int max_histo_idx[5];      // running index
  double max_histo_value[5]; // value(max_histo_idx)
  int most_used_hue[5];      // resulting hue for each channel

  memset(&hue_histo_mean, 0, sizeof(hue_histo_mean));
  memset(&hue_dominance, 0, sizeof(hue_dominance));
  memset(&max_histo_idx, 0, sizeof(max_histo_idx));
  memset(&max_histo_value, 0, sizeof(max_histo_value));
  memset(&most_used_hue, 0, sizeof(most_used_hue));

  // Analyze histogramm for most_used_hue
  for (int channel = 0; channel < 5; ++channel)
  {
    for (int wh_idx = 0; wh_idx < h_MAX + 1; ++wh_idx) // walk trough histogram
    {
      if (w_hue_hist[channel][wh_idx] > max_histo_value[channel])
      {
        // if new value bigger then old one
        max_histo_idx[channel]=wh_idx;            // remember Index
        max_histo_value[channel]=w_hue_hist[channel][wh_idx]; // remember Value
      }
      hue_histo_mean[channel]+=w_hue_hist[channel][wh_idx]; // accumulate histogram values
    }

    hue_histo_mean[channel]=hue_histo_mean[channel] / h_MAX; // calc histogram mean value

    // calc dominace of selected hue value against histogram mean
    // The higher the dominance the higher is the posibility the best hue is selected
    // not used for now but sometimes later
    hue_dominance[channel] = (int)(max_histo_value[channel] / hue_histo_mean[channel]);

    // histogram max index is the most used hue
    most_used_hue[channel]=max_histo_idx[channel];
  }

  // results of Hue histogram analysis are:
  // most_used_hue
  // hue_dominance
  // hue_histo_mean

  /********************************************************************************/
  /*             saturation                                                       */
  /********************************************************************************/

  /*--------------------------------------*/
  /* Saturation Histogram builtup         */
  /*--------------------------------------*/

  double sat_hist[5][s_MAX+1];            // sat histogram
  int pixel_hue = 0;                      // hue of the pixel we are working at
  memset(&sat_hist, 0, sizeof(sat_hist)); // clean histogram

  idx = 0;
  for (int row = 0; row < CAP_HEIGHT; ++row)
  {
    for (int column = 0; column < CAP_WIDTH; ++column)
    {
      // forget black bars:
      // perform calculations only if pixels value is higher than DarkLimit
      if (!(HSV_Img[idx].v < (double)AtmoSetup.DarkLimit * 0.04f))
      {
        // find histogram position for pixel
        hsv_idx = d2i(HSV_Img[idx].s * (double)s_MAX);   // convert from [0;1] -> [0;s_MAX] to use as Histogram Index
        pixel_hue = d2i(HSV_Img[idx].h * (double)h_MAX);

        for (int channel = 0; channel < 5; ++channel)
        {
          // only use pixel for histogram if hue is near most_used_hue
          if ( (pixel_hue > most_used_hue[channel] - 1) && (pixel_hue < most_used_hue[channel] + 1) )
          {
            // build histogram
            sat_hist[channel][hsv_idx]+=Weight[idx].channel[channel]*(HSV_Img[idx].v + 1.0f);
          }
        }
      }
      idx++;
    }
  }

  /*--------------------------------------*/
  /* Saturation Histogram windowing       */
  /*--------------------------------------*/

   double w_sat_hist[5][s_MAX+1];              // windowed HSv histogram
   memset(&w_sat_hist, 0, sizeof(w_sat_hist)); // clean windowed histogram

   // Windowing over Histogram to get most used sat(Hue)
   // TODO implement as function
   // apply lite triangular window design with gradient of 1% per discrete step

   int sat_windowsize = AtmoSetup.SatWinSize; // steps in each direction; eg. 2 => -2 -1 0 1 2 windowing

   // apply windowing to all 5 channels
   for (int channel = 0; channel < 5; ++channel)
   {
     // walk through histogram [0;h_MAX]
     for (int s_idx = 0; s_idx < s_MAX + 1; ++s_idx)
     {
       // windowing from -hue_windowsize -> +hue_windowsize
       for (int mywin = -sat_windowsize; mywin < sat_windowsize + 1; mywin++)
       {
         int myidx = s_idx + mywin; // adressed histogram candlestick

         if (myidx < 0) { myidx = s_MAX + 1 + myidx; } // handle beginning of windowing ->roll back
         if (myidx > h_MAX) { myidx = myidx - s_MAX; } // handle end of windowing -> roll forward

         float windesign = 1 - (0.1 * abs(mywin)); // design of window with 10% gradient

         w_sat_hist[channel][s_idx] += (double)(sat_hist[channel][myidx] * windesign); // apply window
         // w_sat_hist[channel][s_idx]=sat_hist[channel][s_idx]; // apply window
       }
    }
  }

  /*--------------------------------------*/
  /* Analyze Histogram for most used Sat  */
  /*--------------------------------------*/

  // Find Max Vaules in windowed Histogram

  double sat_histo_mean[5]; // Mean Value of Histogram
  int sat_dominance[5];     // Calculate faktor from peak over histogram mean
  int most_used_sat[5];     // resulting sat(most_used_hue) for each channel

  memset(&sat_histo_mean, 0, sizeof(sat_histo_mean));
  memset(&sat_dominance, 0, sizeof(sat_dominance));
  memset(&max_histo_idx, 0, sizeof(max_histo_idx));
  memset(&max_histo_value, 0, sizeof(max_histo_value));
  memset(&most_used_sat, 0, sizeof(most_used_sat));

  // analyze histogramm for most_used_sat
  for (int channel = 0; channel < 5; ++channel)
  {
    for (int wh_idx = 0; wh_idx < s_MAX + 1; ++wh_idx) // walk trough histogram
    {
      if (w_sat_hist[channel][wh_idx] > max_histo_value[channel])
      {
        // if new value bigger then old one
        max_histo_idx[channel] = wh_idx;            // remember Index
        max_histo_value[channel] = w_sat_hist[channel][wh_idx]; // remember Value
      }
      sat_histo_mean[channel] += w_sat_hist[channel][wh_idx];    // accumulate histogram values
    }

    sat_histo_mean[channel] = sat_histo_mean[channel] / (double)(s_MAX); // calc histogram mean value

    // calc dominace of selected hue value against histogram mean
    // The higher the dominance the higher is the posibility the best hue is selected
    // not used for now but sometimes later
    sat_dominance[channel] = (int)(max_histo_value[channel] / sat_histo_mean[channel]);

    // histogram max index is the most used hue
    most_used_sat[channel] = max_histo_idx[channel];
  }

  // results if Hue histogram analysis are:
  // most_used_sat
  // sat_dominance
  // sat_histo_mean

  /*--------------------------------------------------------------*/
  /* find average/peak Value(Brightness) within HSV image         */
  /* uniform Brightness for all channels is calculated            */
  /*--------------------------------------------------------------*/
  int l_counter = 0;
  double value_avg = 0.0f;  // average brightness (value)
  double value_peak = 0.0f; // peak brightness (value)

  // walk through HSV image
  idx = 0;
  for (int row = 0; row < CAP_HEIGHT; ++row )
  {
    for ( int column = 0; column < CAP_WIDTH; ++column)
    {
      // Find value avg/peak
      // only use bright pixels for luminance average
      if (HSV_Img[idx].v > (double)AtmoSetup.DarkLimit * 0.04f)
      {
        // build Brightness average
        value_avg += HSV_Img[idx].v;
        l_counter++;

        // find brightness peak
        if (HSV_Img[idx].v > value_peak) { value_peak = HSV_Img[idx].v; }
      }
      idx++;
    }
  }

  // calculate brightness average
  if (l_counter > 0)
  {
    value_avg = value_avg / (double)(l_counter);
  }
  else
  {
    value_avg = (double)AtmoSetup.DarkLimit * 0.04f;
  }

  tRGBColor rgb_pixel;
  tHSVColor hsv_pixel;
  tColorPacket ColorChannels; // storage container for resulting RGB values

  // copy result of the calculation into ColorChannels
  for (int channel = 0; channel < 5; ++channel)
  {
    hsv_pixel.h = most_used_hue[channel]/(h_MAX*1.0f);
    hsv_pixel.s = most_used_sat[channel]/(s_MAX*1.0f);
    hsv_pixel.v = 1.0f;
    if (AtmoSetup.ValMode == 0) { hsv_pixel.v = value_avg; }
    else if (AtmoSetup.ValMode == 1) { hsv_pixel.v = value_peak; }

    if (AtmoSetup.ValOffset != 0) { hsv_pixel.v *= ((double)AtmoSetup.ValOffset / 10) + 1.0f; } // adjust brightness

    if (hsv_pixel.v > 1.0f) { hsv_pixel.v = 1.0f; } // ensure brightness isnt set too high

    rgb_pixel = HSV2RGB(hsv_pixel); // convert back to rgb

    ColorChannels.channel[channel] = rgb_pixel;
  }

  delete[] Weight;

  return ColorChannels;
}

tHSVColor cInputVideo0::RGB2HSV(tRGBColor color)
{
 double min, max, delta;
 double h = 0.0f, s = 0.0f, v;
 tHSVColor hsv;

 min = MIN(MIN(color.r, color.g), color.b);
 max = MAX(MAX(color.r, color.g), color.b);

 delta = max - min;

 v = max;

 if (delta == 0.0f) // This is a gray, no chroma...
 {
   h = 0.0f;        // HSV results = 0 / 1
   s = 0.0f;
 }
 else // Chromatic data...
 {
   s = delta / max;

   double dr = ( ( ( max - color.r ) / 6.0f ) + ( delta / 2.0f ) ) / delta;
   double dg = ( ( ( max - color.g ) / 6.0f ) + ( delta / 2.0f ) ) / delta;
   double db = ( ( ( max - color.b ) / 6.0f ) + ( delta / 2.0f ) ) / delta;

        if ( color.r == max ) { h = db - dg; }
   else if ( color.g == max ) { h = ( 1.0f / 3.0f ) + dr - db; }
   else if ( color.b == max ) { h = ( 2.0f / 3.0f ) + dg - dr; }

   if ( h < 0.0f ) { h += 1.0f; }
   if ( h > 1.0f ) { h -= 1.0f; }
 }

 hsv.h = h;
 hsv.s = s;
 hsv.v = v;

 return hsv;
}

tRGBColor cInputVideo0::HSV2RGB(tHSVColor color)
{
 tRGBColor rgb = {0.0f, 0.0f, 0.0f};

 float h = (float)color.h*360.0f;
 float s = (float)color.s;
 float v = (float)color.v;

 if (s == 0.0f)
 {
   rgb.r = v;
   rgb.g = rgb.r;
   rgb.b = rgb.r;
 }
 else
 {
   h = h / 60.0f;
   if (h == 6.0f) { h = 0.0f; }
   float i = floor(h);

   float f = h - i;
   float p = v*(1.0f-s);
   float q = v*(1.0f-(s*f));
   float t = v*(1.0f-(s*(1.0f-f)));

   if (i == 0.0f)
   {
     rgb.r = v;
     rgb.g = t;
     rgb.b = p;
   } 
   else if (i == 1.0f)
   {
     rgb.r = q;
     rgb.g = v;
     rgb.b = p;
   }
   else if (i == 2.0f)
   {
     rgb.r = p;
     rgb.g = v;
     rgb.b = t;
   }
   else if (i == 3.0f)
   {
     rgb.r = p;
     rgb.g = q;
     rgb.b = v;
   }
   else if (i == 4.0f)
   {
     rgb.r = t;
     rgb.g = p;
     rgb.b = v;
   }
   else
   {
     rgb.r = v;
     rgb.g = p;
     rgb.b = q;
   }
 }
 return rgb;
}
