/*
 * setup.c: The setup for the Atmolight-plugin
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include "setup.h"

#define SETUP_STARTMODE        "Startmode"
#define SETUP_PERCENTNEW       "FilterSmoothness"
#define SETUP_VALOFFSET        "BrightnessCorrection"
#define SETUP_GAMMAMODE        "GammaMode"
#define SETUP_GAMMAUNI         "GammaUniform"
#define SETUP_GAMMARED         "GammaRed"
#define SETUP_GAMMAGREEN       "GammaGreen"
#define SETUP_GAMMABLUE        "GammaBlue"
#define SETUP_DARKLIMIT        "DarknessLimit"
#define SETUP_MODE             "Mode"
#define SETUP_DEFAULTCOLOR     "DefaultColor"
#define SETUP_STATICRED_CENTER    "SRedC"
#define SETUP_STATICGREEN_CENTER  "SGreenC"
#define SETUP_STATICBLUE_CENTER   "SBlueC"
#define SETUP_STATICRED_LEFT      "SRedL"
#define SETUP_STATICGREEN_LEFT    "SGreenL"
#define SETUP_STATICBLUE_LEFT     "SBlueL"
#define SETUP_STATICRED_RIGHT     "SRedR"
#define SETUP_STATICGREEN_RIGHT   "SGreenR"
#define SETUP_STATICBLUE_RIGHT    "SBlueR"
#define SETUP_STATICRED_TOP       "SRedT"
#define SETUP_STATICGREEN_TOP     "SGreenT"
#define SETUP_STATICBLUE_TOP      "SBlueT"
#define SETUP_STATICRED_BOTTOM    "SRedB"
#define SETUP_STATICGREEN_BOTTOM  "SGreenB"
#define SETUP_STATICBLUE_BOTTOM   "SBlueB"
#define SETUP_CALIBRED            "CalibRed"
#define SETUP_CALIBGREEN          "CalibGreen"
#define SETUP_CALIBBLUE           "CalibBlue"
#define SETUP_HUEWINSIZE          "HueWindowing"
#define SETUP_SATWINSIZE          "SatWindowing"
#define SETUP_VALMODE	          "BrighnessMode"
#define SETUP_FILTERMODE          "TypeOfFilter"
#define SETUP_FILTERMEANCOUNT     "LengthOfFilter"
#define SETUP_FILTERMEANTHRESHOLD "ThresholdOfFilter"
#define SETUP_WEIGHTPOW           "PowerOfWeighting"
#define SETUP_WIDESCREENMODE      "Widescreen Mode"

// seconds the next timer must be in the future to assume manual start
#define MANUALSTART       600

static const char *useStartModeTexts[3];
static const char *useModeTexts[4];
static const char *useGammaModeTexts[3];
static const char *useValueModeTexts[2];
static const char *useFilterTexts[2];
static const char *useDefaultColorsTexts[8];
static const char *useGeneralOnOff[2];

// --- cAtmoSetup ------------------------------------------------------------

cAtmoSetup::cAtmoSetup()
{
  useStartModeTexts[0] = "on";              // start with always on
  useStartModeTexts[1] = "off";             // always off
  useStartModeTexts[2] = "timer dependent"; // timer dependent

  useModeTexts[0] = "live picture (HSV)"; // generate colors out of live picture
  useModeTexts[1] = "default color";      // select default color
  useModeTexts[2] = "static colors";      // generate static color
  useModeTexts[3] = "white calibration";  // user can calibrate generated color to white

  useGammaModeTexts[0] = "off";          // Gamma Correction Off
  useGammaModeTexts[1] = "uniform";      // Uniform Gamma Correction for all channels
  useGammaModeTexts[2] = "by channel";   // Gamma Correction for each RGB channel

  useValueModeTexts[0] = "average";	 // Brightness(value) = average of pic
  useValueModeTexts[1] = "peak";	 //                   = peak of pic

  useFilterTexts[0] = "combined";
  useFilterTexts[1] = "percentage";

  useDefaultColorsTexts[0] = "black";
  useDefaultColorsTexts[1] = "white";
  useDefaultColorsTexts[2] = "red";
  useDefaultColorsTexts[3] = "green";
  useDefaultColorsTexts[4] = "blue";
  useDefaultColorsTexts[5] = "yellow";
  useDefaultColorsTexts[6] = "cyan";
  useDefaultColorsTexts[7] = "magenta";

  useGeneralOnOff[0] = "off";
  useGeneralOnOff[1] = "on";

  output_arg = NULL;

  AtmoInput  = no_input;  // input device
  AtmoOutput = no_output; // output device

  // assumption of manual/automatic start of vdr is done in GetStartMode(),
  // because timers are not loaded yet
  StartMode    = 0;      // start mode = always on
  PercentNew   = 75;     // filter smoothness
  ValOffset    = 0;      // value offset=brightness
  GammaMode    = 0;      // gamma Mode 0=off 1=uniform 2=channel specific
  GammaUni     = 20;     // gamma for uniform mode
  GammaRed     = 20;     // gamma for red channel
  GammaGreen   = 20;     // gamma for green channel
  GammaBlue    = 20;     // gamma for blue channel
  DarkLimit    = 5;      // HSV.V < limit -> HSV.V=0 (no light)
  Mode         = live_picture_hsv; // atmolight-mode
  DefaultColor       = black; // default color
  StaticRed_center   = 255; // static mode: red center channel
  StaticGreen_center = 255; // static mode: green center channel
  StaticBlue_center  = 255; // static mode: blue center channel
  StaticRed_left     = 255; // static mode: red left channel
  StaticGreen_left   = 255; // static mode: green left channel
  StaticBlue_left    = 255; // static mode: blue left channel
  StaticRed_right    = 255; // static mode: red right channel
  StaticGreen_right  = 255; // static mode: green right channel
  StaticBlue_right   = 255; // static mode: blue right channel
  StaticRed_top      = 255; // static mode: red top channel
  StaticGreen_top    = 255; // static mode: green top channel
  StaticBlue_top     = 255; // static mode: blue top channel
  StaticRed_bottom   = 255; // static mode: red bottom channel
  StaticGreen_bottom = 255; // static mode: green bottom channel
  StaticBlue_bottom  = 255; // static mode: blue bottom channel
  CalibRed           = 255; // calibration mode: red
  CalibGreen         = 255; // calibration mode: green
  CalibBlue          = 255; // calibration mode: blue
  HueWinSize          = 5;   // Windowing size for hue histogram building
  SatWinSize          = 5;   // Windowing size for sat histogram building
  ValMode             = 0;   // 0 = avg value, 1 = peak value
  FilterMode          = 0;   // 0 = Combined filter, 1 = Percent filter only
  FilterMeanCount     = 700; // length of floating mean value filter
  FilterMeanThreshold = 70;  // threshold of the mean value filter
  WeightPow           = 12;  // weighting of distance to edge
  WidescreenMode      = 0;   // cut off top and bottom of the pic?
}

bool cAtmoSetup::SetupParse(const char *Name, const char *Value)
{
       if (!strcasecmp(Name, SETUP_STARTMODE))   StartMode   = atoi(Value);
  else if (!strcasecmp(Name, SETUP_PERCENTNEW))  PercentNew  = atoi(Value);
  else if (!strcasecmp(Name, SETUP_VALOFFSET))   ValOffset   = atoi(Value);
  else if (!strcasecmp(Name, SETUP_GAMMAMODE)) 	 GammaMode   = atoi(Value);
  else if (!strcasecmp(Name, SETUP_GAMMAUNI))  	 GammaUni    = atoi(Value);
  else if (!strcasecmp(Name, SETUP_GAMMARED))  	 GammaRed    = atoi(Value);
  else if (!strcasecmp(Name, SETUP_GAMMAGREEN))  GammaGreen  = atoi(Value);
  else if (!strcasecmp(Name, SETUP_GAMMABLUE)) 	 GammaBlue   = atoi(Value);
  else if (!strcasecmp(Name, SETUP_DARKLIMIT))   DarkLimit   = atoi(Value);
  else if (!strcasecmp(Name, SETUP_MODE))        Mode        = (tMode)atoi(Value);
  else if (!strcasecmp(Name, SETUP_DEFAULTCOLOR))           DefaultColor           = (tDefaultColors)atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICRED_CENTER))       StaticRed_center       = atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICGREEN_CENTER))     StaticGreen_center     = atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICBLUE_CENTER))      StaticBlue_center      = atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICRED_LEFT))         StaticRed_left         = atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICGREEN_LEFT))       StaticGreen_left       = atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICBLUE_LEFT))        StaticBlue_left        = atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICRED_RIGHT))        StaticRed_right        = atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICGREEN_RIGHT))      StaticGreen_right      = atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICBLUE_RIGHT))       StaticBlue_right       = atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICRED_TOP))          StaticRed_top          = atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICGREEN_TOP))        StaticGreen_top        = atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICBLUE_TOP))         StaticBlue_top         = atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICRED_BOTTOM))       StaticRed_bottom       = atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICGREEN_BOTTOM))     StaticGreen_bottom     = atoi(Value);
  else if (!strcasecmp(Name, SETUP_STATICBLUE_BOTTOM))      StaticBlue_bottom      = atoi(Value);
  else if (!strcasecmp(Name, SETUP_CALIBRED))               CalibRed    = atoi(Value);
  else if (!strcasecmp(Name, SETUP_CALIBGREEN))             CalibGreen  = atoi(Value);
  else if (!strcasecmp(Name, SETUP_CALIBBLUE))              CalibBlue   = atoi(Value);
  else if (!strcasecmp(Name, SETUP_HUEWINSIZE))          HueWinSize          = atoi(Value);
  else if (!strcasecmp(Name, SETUP_SATWINSIZE))          SatWinSize          = atoi(Value);
  else if (!strcasecmp(Name, SETUP_VALMODE))             ValMode             = atoi(Value);
  else if (!strcasecmp(Name, SETUP_FILTERMODE))          FilterMode          = atoi(Value);
  else if (!strcasecmp(Name, SETUP_FILTERMEANCOUNT))     FilterMeanCount     = atoi(Value);
  else if (!strcasecmp(Name, SETUP_FILTERMEANTHRESHOLD)) FilterMeanThreshold = atoi(Value);
  else if (!strcasecmp(Name, SETUP_WEIGHTPOW))           WeightPow           = atoi(Value);
  else if (!strcasecmp(Name, SETUP_WIDESCREENMODE))      WidescreenMode      = atoi(Value);
  else return false;

  return true;
}

// assumption of manual/automatic start of vdr
void cAtmoSetup::GetStartMode(void)
{
       if (StartMode == 0) { AtmoSetup.atmo_on = true;  } // on
  else if (StartMode == 1) { AtmoSetup.atmo_on = false; } // off
  else if (StartMode == 2) // timer dependent
  {
    time_t Now = time(NULL);
    cTimer *timer = Timers.GetNextActiveTimer();
    time_t Next  = timer ? timer->StartTime() : 0;
    time_t Delta = timer ? Next - Now : 0;

    if (!timer || Delta > MANUALSTART)
    { // Apparently the user started VDR manually
      // dsyslog("Atmolight: assuming manual start of VDR");
      AtmoSetup.atmo_on = true;
    } 
    else
    {
      // dsyslog("Atmolight: assuming automatic start of VDR");
      AtmoSetup.atmo_on = false;
    }
  }
}

void cAtmoSetup::SetDefault(tDefaultColors col)
{
  Mode = default_color;
  DefaultColor = col;
}

void cAtmoSetup::SetStatic(tPosition pos, int r, int g, int b)
{
  Mode = static_color;
  switch (pos)
  {
    default:
    case all:
         StaticRed_center = StaticRed_left = StaticRed_right =
                            StaticRed_top = StaticRed_bottom = r;
         StaticGreen_center = StaticGreen_left = StaticGreen_right =
                              StaticGreen_top = StaticGreen_bottom = g;
         StaticBlue_center = StaticBlue_left = StaticBlue_right =
                             StaticBlue_top = StaticBlue_bottom = b;
    break;

    case center:
         StaticRed_center = r;
         StaticGreen_center = g;
         StaticBlue_center = b;
    break;

    case left:
         StaticRed_left = r;
         StaticGreen_left = g;
         StaticBlue_left = b;
    break;

    case right:
         StaticRed_right = r;
         StaticGreen_right = g;
         StaticBlue_right = b;
    break;

    case top:
         StaticRed_top = r;
         StaticGreen_top = g;
         StaticBlue_top = b;
    break;

    case bottom:
         StaticRed_bottom = r;
         StaticGreen_bottom = g;
         StaticBlue_bottom = b;
    break;
  }
}

// --- cMenuSetupAtmo --------------------------------------------------------

cMenuSetupAtmo::cMenuSetupAtmo(void)
{
  Set();
}

void cMenuSetupAtmo::Set(void)
{
  int current = Current();
  Clear();

  Add(new cMenuEditStraItem("Startmode", &AtmoSetup.StartMode, 3, useStartModeTexts));

  Add(new cMenuEditStraItem("Widescreen mode", &AtmoSetup.WidescreenMode, 2, useGeneralOnOff));

  Add(new cMenuEditIntItem("Power of weighting", &AtmoSetup.WeightPow, 1, 30)); 

  Add(new cMenuEditIntItem("Brightness", &AtmoSetup.ValOffset, -10, 30));

  Add(new cMenuEditStraItem("Gamma mode", &AtmoSetup.GammaMode, 3, useGammaModeTexts));
  if (AtmoSetup.GammaMode == 1)
  {
    Add(new cMenuEditIntItem("  Gamma" , &AtmoSetup.GammaUni, 10, 50));
  }
  else if (AtmoSetup.GammaMode == 2)
  {
    Add(new cMenuEditIntItem("  GammaRed" ,   &AtmoSetup.GammaRed,   10, 50));
    Add(new cMenuEditIntItem("  GammaGreen" , &AtmoSetup.GammaGreen, 10, 50));
    Add(new cMenuEditIntItem("  GammaBlue" ,  &AtmoSetup.GammaBlue,  10, 50));
  }

  Add(new cMenuEditStraItem("Filter", &AtmoSetup.FilterMode, 2, useFilterTexts));
  if (AtmoSetup.FilterMode == 0)
  {
   Add(new cMenuEditIntItem("  Filter length [ms]", &AtmoSetup.FilterMeanCount, 300, 5000));
   Add(new cMenuEditIntItem("  Filter threshold [%]", &AtmoSetup.FilterMeanThreshold, 1, 100));
   Add(new cMenuEditIntItem("  Filter smoothness [%]", &AtmoSetup.PercentNew, 1, 100));
  }
  else if (AtmoSetup.FilterMode == 1)
  {
   Add(new cMenuEditIntItem("  Filter smoothness [%]", &AtmoSetup.PercentNew, 1, 100));
  }

  Add(new cMenuEditStraItem("Mode", (int*)&AtmoSetup.Mode, 4, useModeTexts));
  if (AtmoSetup.Mode == live_picture_hsv)
  {
    Add(new cMenuEditIntItem("  Darkness limit", &AtmoSetup.DarkLimit, 0, 10));
    Add(new cMenuEditIntItem("  Hue windowing", &AtmoSetup.HueWinSize, 0, 5));
    Add(new cMenuEditIntItem("  Sat windowing", &AtmoSetup.SatWinSize, 0, 5));
    Add(new cMenuEditStraItem("  Brightness mode", &AtmoSetup.ValMode, 2, useValueModeTexts));
  }
  else if (AtmoSetup.Mode == default_color)
  {
    Add(new cMenuEditStraItem("  Default Color", (int*)&AtmoSetup.DefaultColor, 8, useDefaultColorsTexts));
  }
  else if (AtmoSetup.Mode == static_color)
  {
    Add(new cMenuEditIntItem("  Red center"  , &AtmoSetup.StaticRed_center, 0, 255));
    Add(new cMenuEditIntItem("  Green center", &AtmoSetup.StaticGreen_center, 0, 255));
    Add(new cMenuEditIntItem("  Blue center" , &AtmoSetup.StaticBlue_center, 0, 255));
    Add(new cMenuEditIntItem("  Red left"    , &AtmoSetup.StaticRed_left, 0, 255));
    Add(new cMenuEditIntItem("  Green left"  , &AtmoSetup.StaticGreen_left, 0, 255));
    Add(new cMenuEditIntItem("  Blue left"   , &AtmoSetup.StaticBlue_left, 0, 255));
    Add(new cMenuEditIntItem("  Red right"   , &AtmoSetup.StaticRed_right, 0, 255));
    Add(new cMenuEditIntItem("  Green right" , &AtmoSetup.StaticGreen_right, 0, 255));
    Add(new cMenuEditIntItem("  Blue right"  , &AtmoSetup.StaticBlue_right, 0, 255));
    Add(new cMenuEditIntItem("  Red top"     , &AtmoSetup.StaticRed_top, 0, 255));
    Add(new cMenuEditIntItem("  Green top"   , &AtmoSetup.StaticGreen_top, 0, 255));
    Add(new cMenuEditIntItem("  Blue top"    , &AtmoSetup.StaticBlue_top, 0, 255));
    Add(new cMenuEditIntItem("  Red bottom"  , &AtmoSetup.StaticRed_bottom, 0, 255));
    Add(new cMenuEditIntItem("  Green bottom", &AtmoSetup.StaticGreen_bottom, 0, 255));
    Add(new cMenuEditIntItem("  Blue bottom" , &AtmoSetup.StaticBlue_bottom, 0, 255));
  }
  else if (AtmoSetup.Mode == white_calibration)
  {
    Add(new cMenuEditIntItem("  Red"  , &AtmoSetup.CalibRed, 0, 255));
    Add(new cMenuEditIntItem("  Green", &AtmoSetup.CalibGreen, 0, 255));
    Add(new cMenuEditIntItem("  Blue" , &AtmoSetup.CalibBlue, 0, 255));
  }

  SetCurrent(Get(current));
  Display();
}

void cMenuSetupAtmo::Store(void)
{
  SetupStore(SETUP_STARTMODE,   AtmoSetup.StartMode);
  SetupStore(SETUP_PERCENTNEW,  AtmoSetup.PercentNew);
  SetupStore(SETUP_VALOFFSET,   AtmoSetup.ValOffset);
  SetupStore(SETUP_GAMMAMODE,   AtmoSetup.GammaMode);
  SetupStore(SETUP_GAMMAUNI,    AtmoSetup.GammaUni);
  SetupStore(SETUP_GAMMARED,    AtmoSetup.GammaRed);
  SetupStore(SETUP_GAMMAGREEN,  AtmoSetup.GammaGreen);
  SetupStore(SETUP_GAMMABLUE,   AtmoSetup.GammaBlue);
  SetupStore(SETUP_DARKLIMIT,   AtmoSetup.DarkLimit);
  SetupStore(SETUP_MODE,        AtmoSetup.Mode);
  SetupStore(SETUP_DEFAULTCOLOR,          AtmoSetup.DefaultColor);
  SetupStore(SETUP_STATICRED_CENTER,      AtmoSetup.StaticRed_center);
  SetupStore(SETUP_STATICGREEN_CENTER,    AtmoSetup.StaticGreen_center);
  SetupStore(SETUP_STATICBLUE_CENTER,     AtmoSetup.StaticBlue_center);
  SetupStore(SETUP_STATICRED_LEFT,        AtmoSetup.StaticRed_left);
  SetupStore(SETUP_STATICGREEN_LEFT,      AtmoSetup.StaticGreen_left);
  SetupStore(SETUP_STATICBLUE_LEFT,       AtmoSetup.StaticBlue_left);
  SetupStore(SETUP_STATICRED_RIGHT,       AtmoSetup.StaticRed_right);
  SetupStore(SETUP_STATICGREEN_RIGHT,     AtmoSetup.StaticGreen_right);
  SetupStore(SETUP_STATICBLUE_RIGHT,      AtmoSetup.StaticBlue_right);
  SetupStore(SETUP_STATICRED_TOP,         AtmoSetup.StaticRed_top);
  SetupStore(SETUP_STATICGREEN_TOP,       AtmoSetup.StaticGreen_top);
  SetupStore(SETUP_STATICBLUE_TOP,        AtmoSetup.StaticBlue_top);
  SetupStore(SETUP_STATICRED_BOTTOM,      AtmoSetup.StaticRed_bottom);
  SetupStore(SETUP_STATICGREEN_BOTTOM,    AtmoSetup.StaticGreen_bottom);
  SetupStore(SETUP_STATICBLUE_BOTTOM,     AtmoSetup.StaticBlue_bottom);
  SetupStore(SETUP_CALIBRED,              AtmoSetup.CalibRed);
  SetupStore(SETUP_CALIBGREEN,            AtmoSetup.CalibGreen);
  SetupStore(SETUP_CALIBBLUE,             AtmoSetup.CalibBlue);
  SetupStore(SETUP_HUEWINSIZE,         AtmoSetup.HueWinSize);
  SetupStore(SETUP_SATWINSIZE,         AtmoSetup.SatWinSize);
  SetupStore(SETUP_VALMODE,            AtmoSetup.ValMode);
  SetupStore(SETUP_FILTERMODE,         AtmoSetup.FilterMode);
  SetupStore(SETUP_FILTERMEANCOUNT,    AtmoSetup.FilterMeanCount);
  SetupStore(SETUP_FILTERMEANTHRESHOLD,AtmoSetup.FilterMeanThreshold);
  SetupStore(SETUP_WEIGHTPOW,          AtmoSetup.WeightPow);
  SetupStore(SETUP_WIDESCREENMODE,     AtmoSetup.WidescreenMode);
}

eOSState cMenuSetupAtmo::ProcessKey(eKeys Key)
{
  int  prevMode       = AtmoSetup.Mode;
  int  prevGammaMode  = AtmoSetup.GammaMode;
  int  prevFilterMode = AtmoSetup.FilterMode;

  eOSState state = cOsdMenu::ProcessKey(Key);

  if( (prevMode != AtmoSetup.Mode) || 
      (prevGammaMode != AtmoSetup.GammaMode) || 
      (prevFilterMode != AtmoSetup.FilterMode) )
  {
    Set();
    state = osContinue;
  }

  if (state == osUnknown)
  {
    switch (Key)
    {
      case kOk: 
           Store();
           state = osBack;
      break;

      default:
      break;
    }
  }
  return state;
}
