/*
 * thread.c: The thread for the Atmolight-plugin
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */
#include <stdlib.h>
#include <vdr/skins.h>
#include <vdr/tools.h>
#include <vdr/thread.h>
#include "math.h"

#include "defs.h"

// filter class
#include "filter.h"

// input classes
#include "input.h"
#include "inputvideo0.h"

// output classes
#include "output.h"
#include "outputserial.h"
#include "outputnetwork.h"

#include "setup.h"
#include "thread.h"

//************************************************************************************************************
//***             cAtmoThread                                                                              ***
//************************************************************************************************************

cAtmoThread::cAtmoThread()
{
  // input device
  switch (AtmoSetup.AtmoInput)
  {
    case video0:
         Input = new cInputVideo0();
    break;

    default:
    case no_input:
         Input = NULL;
         Skins.QueueMessage(mtInfo, "Atmolight: No input device given!"); // TODO: tr("")
    break;
  }

  // filter
  Filter = new cAtmoFilter();

  // output device
  switch (AtmoSetup.AtmoOutput)
  {
    case serial:
         Output = new cOutputSerial();
    break;

    case network:
         Output = new cOutputNetwork();
    break;

    default:
    case no_output:
         Output = NULL;
         Skins.QueueMessage(mtInfo, "Atmolight: No output device given!"); // TODO: tr("")
    break;
  }
}

cAtmoThread::~cAtmoThread()
{  
  AtmoSetup.atmo_on = false;
  OutputColors(ColorPacketInt); // send all off

  if (Input)
  {
    Input->Close();
    delete Input;
    Input = NULL;
  }

  delete Filter;
  Filter = NULL;

  if (Output)
  {
    Output->Close();
    delete Output;
    Output = NULL;
  }
}

void cAtmoThread::SwitchOn(void)
{
  AtmoSetup.atmo_on = true;

#if 0
  if (false == Input->Open(NULL))
  {
    AtmoSetup.atmo_on = false;
    AtmoSetup.AtmoInput = no_input;
    Skins.QueueMessage(mtInfo, "Atmolight: Could not open input device!"); // TODO: tr("")
    return;
  }
  if (false == Output->Open(AtmoSetup.output_arg))
  {
    AtmoSetup.atmo_on = false;
    AtmoSetup.AtmoOutput = no_output;
    Skins.QueueMessage(mtInfo, "Atmolight: Could not open output device!"); // TODO: tr("")
  }
#endif
}

void cAtmoThread::SwitchOff(void)
{
  AtmoSetup.atmo_on = false;
  OutputColors(ColorPacketInt); // send all off
#if 0
  Input->Close();
  Output->Close();
#endif
}

void cAtmoThread::Action(void)
{
  cTimeMs t;

  if ((NULL == Input) || (NULL == Output)) { return; } // input and output existing?

  cCondWait::SleepMs(1000); // wait 1,0 second // TODO: verkleinern

  AtmoSetup.GetStartMode(); // manual or automatic start of vdr?

  if (AtmoSetup.atmo_on)
  {
    /* SwitchOn(); */
    if (false == Input->Open(NULL))
    {
      AtmoSetup.atmo_on = false;
      AtmoSetup.AtmoInput = no_input;
      Skins.QueueMessage(mtInfo, "Atmolight: Could not open input device!"); // TODO: tr("")
      return;
    }
    if (false == Output->Open(AtmoSetup.output_arg))
    {
      AtmoSetup.atmo_on = false;
      AtmoSetup.AtmoOutput = no_output;
      Skins.QueueMessage(mtInfo, "Atmolight: Could not open output device!"); // TODO: tr("")
    }
  }

  while (Running())
  {
    t.Set(0);

    if (AtmoSetup.atmo_on)
    {
      // get calculated color from input
      ColorPacketInt = Input->GetColorPacket();

      // send to filter
      ColorPacketInt = Filter->Filtering(ColorPacketInt);

      // apply gamma
      ColorPacketInt = ApplyGamma(ColorPacketInt);

      // apply white calibration
      ColorPacketInt = WhiteCalibration(ColorPacketInt);

      // send to output device
      OutputColors(ColorPacketInt);
    }

    //esyslog("%d",t.Elapsed()); // print duration
    if (t.Elapsed() < 20) // ensure that the filter is called every 20ms
    {
      cCondWait::SleepMs(20 - t.Elapsed());
    }
  }
}

// apply gamma
tColorPacketInt cAtmoThread::ApplyGamma(tColorPacketInt ColorPacketInt)
{
  if (AtmoSetup.GammaMode == 1)
  {
    // uniform mode: all channels same gamma
    double GammaUni = (double)AtmoSetup.GammaUni;

    for (int i = 0; i < 5; i++)
    {
      ColorPacketInt.channel[i].r = (unsigned char)(pow((double)ColorPacketInt.channel[i].r / 255.0f, GammaUni / 10.0f) * 255.0f);
      ColorPacketInt.channel[i].g = (unsigned char)(pow((double)ColorPacketInt.channel[i].g / 255.0f, GammaUni / 10.0f) * 255.0f);
      ColorPacketInt.channel[i].b = (unsigned char)(pow((double)ColorPacketInt.channel[i].b / 255.0f, GammaUni / 10.0f) * 255.0f);
    }
  }
  else if (AtmoSetup.GammaMode == 2)
  {
    // gamma correction for each channel
    double GammaRed   = (double)AtmoSetup.GammaRed;
    double GammaGreen = (double)AtmoSetup.GammaGreen;
    double GammaBlue  = (double)AtmoSetup.GammaBlue;

    for (int i = 0; i < 5; i++)
    {
      ColorPacketInt.channel[i].r = (unsigned char)(pow((double)ColorPacketInt.channel[i].r / 255.0f, GammaRed   / 10.0f) * 255.0f);
      ColorPacketInt.channel[i].g = (unsigned char)(pow((double)ColorPacketInt.channel[i].g / 255.0f, GammaGreen / 10.0f) * 255.0f);
      ColorPacketInt.channel[i].b = (unsigned char)(pow((double)ColorPacketInt.channel[i].b / 255.0f, GammaBlue  / 10.0f) * 255.0f);
    }
  }
  return ColorPacketInt;
}

// apply white calibration
tColorPacketInt cAtmoThread::WhiteCalibration(tColorPacketInt ColorPacketInt)
{
  for (int i = 0; i < 5; i++)
  {
    // ColorPacket.channel[i].r = ((float)AtmoSetup.CalibRed   / 255.0f) * ColorPacket.channel[i].r;
    // ColorPacket.channel[i].g = ((float)AtmoSetup.CalibGreen / 255.0f) * ColorPacket.channel[i].g;
    // ColorPacket.channel[i].b = ((float)AtmoSetup.CalibBlue  / 255.0f) * ColorPacket.channel[i].b;
    ColorPacketInt.channel[i].r = (unsigned char)((int)AtmoSetup.CalibRed   * (int)ColorPacketInt.channel[i].r / 255);
    ColorPacketInt.channel[i].g = (unsigned char)((int)AtmoSetup.CalibGreen * (int)ColorPacketInt.channel[i].g / 255);
    ColorPacketInt.channel[i].b = (unsigned char)((int)AtmoSetup.CalibBlue  * (int)ColorPacketInt.channel[i].b / 255);
  }

  return ColorPacketInt;
}

// write the filter-output to the device
void cAtmoThread::OutputColors(const tColorPacketInt ColorPacketInt)
{
  tColorPacketInt col;
  static tColorPacketInt last_col;

  if (AtmoSetup.atmo_on)
  {
    switch (AtmoSetup.Mode)
    {
      case live_picture_hsv: // live mode
           col = ColorPacketInt;
      break;

      case default_color: // default color
           switch (AtmoSetup.DefaultColor)
           {
             case black:
             default:
                  for (int i=0; i<5; i++)
                  {
                    col.channel[i].r = 0; // red (0-255)
                    col.channel[i].g = 0; // green (0-255)
                    col.channel[i].b = 0; // blue (0-255)
                  }
             break;

             case white:
                  for (int i=0; i<5; i++)
                  {
                    col.channel[i].r = 255; // red (0-255)
                    col.channel[i].g = 255; // green (0-255)
                    col.channel[i].b = 255; // blue (0-255)
                  }
             break;

             case red:
                  for (int i=0; i<5; i++)
                  {
                    col.channel[i].r = 255; // red (0-255)
                    col.channel[i].g = 0;   // green (0-255)
                    col.channel[i].b = 0;   // blue (0-255)
                  }
             break;

             case green:
                  for (int i=0; i<5; i++)
                  {
                    col.channel[i].r = 0;   // red (0-255)
                    col.channel[i].g = 255; // green (0-255)
                    col.channel[i].b = 0;   // blue (0-255)
                  }
             break;

             case blue:
                  for (int i=0; i<5; i++)
                  {
                    col.channel[i].r = 0;   // red (0-255)
                    col.channel[i].g = 0;   // green (0-255)
                    col.channel[i].b = 255; // blue (0-255)
                  }
             break;

             case yellow:
                  for (int i=0; i<5; i++)
                  {
                    col.channel[i].r = 255; // red (0-255)
                    col.channel[i].g = 255; // green (0-255)
                    col.channel[i].b = 0;   // blue (0-255)
                  }
             break;

             case cyan:
                  for (int i=0; i<5; i++)
                  {
                    col.channel[i].r = 0;   // red (0-255)
                    col.channel[i].g = 255; // green (0-255)
                    col.channel[i].b = 255; // blue (0-255)
                  }
             break;

             case magenta:
                  for (int i=0; i<5; i++)
                  {
                    col.channel[i].r = 255; // red (0-255)
                    col.channel[i].g = 0;   // green (0-255)
                    col.channel[i].b = 255; // blue (0-255)
                  }
             break;
           }
      break;

      case static_color: // static mode
           col.channel[0].r = AtmoSetup.StaticRed_center;   // red (0-255)
           col.channel[0].g = AtmoSetup.StaticGreen_center; // green (0-255)
           col.channel[0].b = AtmoSetup.StaticBlue_center;  // blue (0-255)

           col.channel[1].r = AtmoSetup.StaticRed_left;
           col.channel[1].g = AtmoSetup.StaticGreen_left;
           col.channel[1].b = AtmoSetup.StaticBlue_left;

           col.channel[2].r = AtmoSetup.StaticRed_right;
           col.channel[2].g = AtmoSetup.StaticGreen_right;
           col.channel[2].b = AtmoSetup.StaticBlue_right;

           col.channel[3].r = AtmoSetup.StaticRed_top;
           col.channel[3].g = AtmoSetup.StaticGreen_top;
           col.channel[3].b = AtmoSetup.StaticBlue_top;

           col.channel[4].r = AtmoSetup.StaticRed_bottom;
           col.channel[4].g = AtmoSetup.StaticGreen_bottom;
           col.channel[4].b = AtmoSetup.StaticBlue_bottom;
      break;

      case white_calibration: // white calibration
           // TODO: different white calibration for each channel???
           col.channel[0].r = AtmoSetup.CalibRed;   // red (0-255)
           col.channel[0].g = AtmoSetup.CalibGreen; // green (0-255)
           col.channel[0].b = AtmoSetup.CalibBlue;  // blue (0-255)

           col.channel[1].r = AtmoSetup.CalibRed;
           col.channel[1].g = AtmoSetup.CalibGreen;
           col.channel[1].b = AtmoSetup.CalibBlue;

           col.channel[2].r = AtmoSetup.CalibRed;
           col.channel[2].g = AtmoSetup.CalibGreen;
           col.channel[2].b = AtmoSetup.CalibBlue;

           col.channel[3].r = AtmoSetup.CalibRed;
           col.channel[3].g = AtmoSetup.CalibGreen;
           col.channel[3].b = AtmoSetup.CalibBlue;

           col.channel[4].r = AtmoSetup.CalibRed;
           col.channel[4].g = AtmoSetup.CalibGreen;
           col.channel[4].b = AtmoSetup.CalibBlue;
      break;
    }
  }
  else // (false == AtmoSetup.atmo_on)
  {
    col.channel[0].r = col.channel[0].g = col.channel[0].b = 0;
    col.channel[4] = col.channel[3] = col.channel[2] = col.channel[1] = col.channel[0];
  }

  if (Output)
  {
    if (0 != memcmp(&last_col, &col, sizeof(last_col)))
    {
      Output->OutputColors(col);
    }
    memcpy(&last_col, &col, sizeof(last_col));
  }
}
